{
****************************************************************
*  Name    : ISRRX2 Module                                     *
*  Author  : David John Barker                                 *
*  Notice  : Copyright (c) 2006 Mecanique                      *
*          : All Rights Reserved                               *
*  Date    : 30/12/2020                                        *
*  Version : 1.2                                               *
*  Notes   : 1.1 - initial port from ISRRX                     *
*          : 1.2 - add '#option RX2_USART_SELECT' to enable    *
*          :       support for USART1-USART5                   *
*          : 1.3 - add '#option RX2_USER_ISR' to allow using   *
*          :       a user-defined interrupt() routine          *
****************************************************************
}
module ISRRX2

#option RX2_USART_SELECT = 1
#if not(RX2_USART_SELECT in (1, 2, 3, 4, 5))
   #error RX2_USART_SELECT, "Invalid option. ISRRX2 supports USART 1-5"
#endif
// verify device supports USARTx
#if (RX2_USART_SELECT > _usart)
   #error RX2_USART_SELECT, "Invalid option. Device does not support selected USART"
#endif

// import USARTx library...
#if (RX2_USART_SELECT = 1)
  include "USART.bas"
#elseif (RX2_USART_SELECT = 2)
  include "USART2.bas"
#elseif (RX2_USART_SELECT = 3)
  include "USART3.bas"
#elseif (RX2_USART_SELECT = 4)
  include "USART4.bas"
#elseif (RX2_USART_SELECT = 5)
  include "USART5.bas"
#endif
public dim USART_OVERRUN as Overrun    // alias for USARTx.Overrun status flag

// set interrupt priority level...
#option RX2_PRIORITY = ipHigh
#if not(RX2_PRIORITY in (ipLow, ipHigh))
   #error RX2_PRIORITY, "Invalid option. Priority must be ipHigh or ipLow"
#endif
public const PriorityLevel = RX2_PRIORITY

// size of the RX buffer...
#option RX2_BUFFER_SIZE = 64      
#if not(RX2_BUFFER_SIZE in (1 to 255))
   #error RX2_BUFFER_SIZE, "Invalid option. Buffer size must be between 1 and 255"
#endif
public const BufferSize = RX2_BUFFER_SIZE

// select interrupt handler isr (local routine or user-defined)
//  false = local OnRX interrupt, default 
//  true  = user-defined interrupt (external)
// if using an external interrupt routine it is up to the user to define an 'interrupt()' handler
// and call 'enable()'
#option RX2_USER_ISR = false

type TEvent = event()

// local variables and aliases...
dim 
   FBuffer(BufferSize) as byte,
   FIndexIn as byte,
   FIndexOut as byte,
   FByteRead as byte,
   FProcessByte as boolean,
   FMaybeOverrun as boolean,
   FOnDataEvent as TEvent
 
// public variables and aliases...   
public dim   
   USARTOverrun as USART_OVERRUN,     // USARTx.Overrun status
   BufferOverrun as boolean,
   DataByte as FByteRead,
   DataChar as FByteRead.AsChar,
   ProcessByte as FProcessByte

{
****************************************************************************
* interrupt helper functions                                               *
****************************************************************************
}
public inline sub EnableIntr()
    RCIEnable = true            // USARTx.RCIEnable
end sub

public inline sub DisableIntr()
    RCIEnable = false           // USARTx.RCIEnable
end sub

{
****************************************************************************
* Name    : RX_intr_handler                                                *
* Purpose : ISR routine to buffer incoming data                            *
****************************************************************************
}
// intr buffering routine (can be used by external isr)
dim FSRSave as word
public inline sub RX_intr_handler()
   FSRSave = FSR0
   BufferOverrun = FMaybeOverrun
   if not(USART_OVERRUN) then       // USARTx.Overrun
      FByteRead = RCRegister        // USARTx.RCRegister
      if not(BufferOverrun) then
         FProcessByte = true
         
         // execute handler...
         FOnDataEvent()
         
         if (FProcessByte) then
            FBuffer(FIndexIn) = FByteRead
            inc(FIndexIn)
            if (FIndexIn > bound(FBuffer)) then
               FIndexIn = 0
            endif
            FMaybeOverrun = (FIndexIn = FIndexOut) 
         endif   
      endif
   else                 // usart overrun...
      DisableIntr()     // disable further rx intr until handled by user
   endif         
   FSR0 = FSRSave
end sub

{
****************************************************************************
* Name    : OnRX (PRIVATE)                                                 *
* Purpose : Interrupt Service Routine (ISR) to buffer incoming data        *
****************************************************************************
}
// local intr ISR (default)
interrupt OnRX(PriorityLevel)
    Rx_intr_handler()
end interrupt

{
****************************************************************************
* Name    : Initialize                                                     *
* Purpose : Initialize buffering - with optional OnData event handler      *
****************************************************************************
}
public sub Initialize(pOnDataEvent as TEvent = 0, EnaIntr as boolean = true)
   DisableIntr()
   FOnDataEvent = pOnDataEvent
   FIndexIn = 0
   FIndexOut = 0
   FMaybeOverrun = false
   BufferOverrun = false

   ClearOverrun()               // USARTx.ClearOverrun
  #if (RX2_PRIORITY = ipHigh)
   RCIPHigh = true              // USARTx.RCIPHigh
  #else
   RCIPHigh = false             // USARTx.RCIPHigh
  #endif
   if (EnaIntr) then
      EnableIntr()
   endif      
  #if (RX2_USER_ISR = false)
   enable(OnRX)
  #endif
end sub

{
****************************************************************************
* Name    : Reset                                                          *
* Purpose : Reset module                                                   *
****************************************************************************
}
public sub Reset()
   DisableIntr()
   Initialize(FOnDataEvent, true)
end sub   

{
****************************************************************************
* Name    : Start                                                          *
* Purpose : Start interrupt handling                                       *
****************************************************************************
}
public sub Start()
   EnableIntr()
  #if (RX2_USER_ISR = false)
   enable(OnRX)
  #endif
end sub

{
****************************************************************************
* Name    : Stop                                                           *
* Purpose : Stop interrupt handling                                        *
****************************************************************************
}
public sub Stop()
   DisableIntr()
end sub

{
****************************************************************************
* Name    : DataAvailable                                                  *
* Purpose : Check to see if there is data in the buffer                    *
****************************************************************************
}
public function DataAvailable() as boolean
   DisableIntr()
   Result = FIndexIn <> FIndexOut
   EnableIntr()
end function

{
****************************************************************************
* Name    : Overrun                                                        *
* Purpose : Returns true if RC register or buffer has overrun, false       *
*         : otherwise                                                      *
****************************************************************************
}
public function Overrun() as boolean
   Result = USART_OVERRUN or BufferOverrun   // USARTx.Overrun
end function

{
****************************************************************************
* Name    : GetByte (PRIVATE)                                              *
* Purpose : Get a single byte from the buffer                              *
****************************************************************************
}
function GetByte() as byte
   FMaybeOverrun = false
   Result = FBuffer(FIndexOut)
   inc(FIndexOut)
   if (FIndexOut > bound(FBuffer)) then
      FIndexOut = 0
   endif   
end function

{
****************************************************************************
* Name    : ReadByte                                                       *
* Purpose : Read a single byte from the buffer                             *
****************************************************************************
}
public function ReadByte() as byte
   DisableIntr()
   Result = GetByte()
   EnableIntr()
end function

{
****************************************************************************
* Name    : ReadWord                                                       *
* Purpose : Read a word from the buffer                                    *
****************************************************************************
}
public function ReadWord() as word
   DisableIntr()
   Result.Bytes(0) = GetByte()
   Result.Bytes(1) = GetByte()
   EnableIntr()
end function

{
****************************************************************************
* Name    : ReadLongWord                                                   *
* Purpose : Read a long word from the buffer                               *
****************************************************************************
}
public function ReadLongWord() as longword
   DisableIntr()
   Result.Bytes(0) = GetByte()
   Result.Bytes(1) = GetByte()
   Result.Bytes(2) = GetByte()
   Result.Bytes(3) = GetByte()
   EnableIntr()
end function

{
****************************************************************************
* Name    : ReadFloat                                                      *
* Purpose : Read a floating point number from the buffer                   *
****************************************************************************
}
public function ReadFloat() as float
   DisableIntr()
   Result.Bytes(0) = GetByte()
   Result.Bytes(1) = GetByte()
   Result.Bytes(2) = GetByte()
   Result.Bytes(3) = GetByte()
   EnableIntr()
end function

{
****************************************************************************
* Name    : ReadStr                                                        *
* Purpose : Read a string from the buffer. Optional parameter pTerminator  *
*         : to specify the input string terminator character. The function *
*         : returns the number of characters read                          *
****************************************************************************
}
public function ReadStr(byref pText as string, pTerminator as char = null) as byte
   dim ch as char
   
   DisableIntr()
   Result = 0
   FSR0 = addressof(pText)
   repeat
      ch = GetByte()
      if (ch <> pTerminator) then
         POSTINC0 = ch
         inc(Result)
      endif   
   until (ch = pTerminator)
   POSTINC0 = 0
   EnableIntr()
end function

//
// module init
//
// to satisfy 'might not have been initialized' warnings
BufferOverrun = false
FBuffer(0) = 0

end module
